//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gertespezifischer Funktionen und Mechanismen            //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright (c) 2010 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include "AVRChip.h"
#include "Conf.h"
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Aufgabenzuordnung der Port-Pins  /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define SYSTEM_LED      A,7
#define DISPLAY_BL      C,4
#define DISPLAY_EN      C,5
#define DISPLAY_RW      C,6
#define DISPLAY_RS      C,7
#define RELAY_SET       D,4
#define RELAY2_EN       D,5
#define RELAY1_EN       D,6

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum
{   DO_CONTINUE,
    DO_RESTART,
    DO_ENTER_APPL
} DoNext;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt16           prgLdrMode __attribute__ ((section (".noinit")));
static volatile UInt16  msecTicks                               =   0;
static volatile UInt08  busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY+1] = { 0 };
static volatile UInt08  busRxQueueHead                          =   0;
static volatile UInt08  busRxQueueTail                          =   0;
static volatile UInt08  serRxQueue[CFG_SERIAL_RX_QUEUE_CAPACITY+1] = { 0 };
static volatile UInt08  serRxQueueHead                          =   0;
static volatile UInt08  serRxQueueTail                          =   0;
static          DoNext  doNext                                  = DO_CONTINUE;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void __attribute__((noreturn))
Restart(void)
{
    DisableInterrupts();
    prgLdrMode = CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while (true);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void __attribute__((noreturn))
EnterAppl(void)
{
    DisableInterrupts();
    prgLdrMode = !CFG_PGMLDR_MARKER;
    wdt_enable(WDTO_15MS);
    while (true);
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt08
DisplayRead(bool regSet)
{
    UInt08 data;

    ConfPort(B, _T_, _T_, _T_, _T_, _T_, _T_, _T_, _T_); // Datenleitungen auf Eingang
    Atomic
    {   ChgPin(DISPLAY_RS, regSet);
        SetPin(DISPLAY_RW);
        SetPin(DISPLAY_EN);
        asm volatile("nop");
        data = PINB;
        ClrPin(DISPLAY_EN);
    }
    return data;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
DisplayWrite(bool regSet, UInt08 data)
{
    ConfPort(B, _L_, _L_, _L_, _L_, _L_, _L_, _L_, _L_); // Datenleitungen auf Ausgang
    PORTB = data;
    Atomic
    {   ChgPin(DISPLAY_RS, regSet);
        ClrPin(DISPLAY_RW);
        SetPin(DISPLAY_EN);
        asm volatile("nop");
        ClrPin(DISPLAY_EN);
    }
    // Warten bis Vorgang abgeschlossen  ist (Busy Flag)
    while (DisplayRead(false) & BV(7));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
DisplayEnable(void)
{
    DisplayWrite(false, 0b00111000); // 8-Bit Datenbreite, 2-Zeilen, 5x7 Zeichensatz
    DisplayWrite(false, 0b00001100); // Anzeige ein, Cursor unsichtbar, kein blinken
    DisplayWrite(false, 0b00000001); // Anzeige lschen, Cursor links/oben
    DisplayWrite(false, 0b00000110); // Autoinkrementierung der Adresse
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedRelaySwitchOff(void)
{
    static UInt16  timeout = 0;
    static UInt08  stage   = 0;

    if (stage != 0 && !BOARD_Timeout(timeout)) return;
    switch( stage )
    {   case 0:
            SetPin(RELAY_SET);  // true = Relais aus
            break;
        case 1:
            SetPin(RELAY1_EN);
            break;
        case 2:
            ClrPin(RELAY1_EN);
            break;
        case 3:
            SetPin(RELAY2_EN);
            break;
        case 4:
            ClrPin(RELAY2_EN);
            break;
        default:
            return;
    }
    stage++;
    timeout = BOARD_SetTimeout(CFG_RELAY_SWITCH_STEP_TIME);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Vorlauf  /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//                          (X)                                                         //
//                           |                                                          //
//                           |                                                          //
//                          \|/                                                         //
//                    /\  ja                                              //
//                   | Power-On-Reset |--------------------+                            //
//                    \______________/                     |                            //
//                           | nein                        |                            //
//                           |                             |                            //
//                          \|/                           \|/                           //
//                    /\ nein         /\ ja                 //
//                   | External-Reset |----------->| Watchdog-Reset |------+            //
//                    \______________/              \______________/       |            //
//                           | ja                          | nein          |            //
//                           |                             |               |            //
//                          \|/                           \|/              |            //
//                +----------------------+     +-----------------------+   |            //
//                | mode = PGMLDR_MARKER |     | mode = !PGMLDR_MARKER |   |            //
//                +----------------------+     +-----------------------+   |            //
//                           |                             |               |            //
//                           |<----------------------------+---------------+            //
//                           |                                                          //
//                          \|/                                                         //
//                /\  nein                                         //
//               | mode != PGMLDR_MARKER |-----------------+                            //
//                \_____________________/                  |                            //
//                           | ja                          |                            //
//                           |                             |                            //
//                          \|/                            |                            //
//                  /\  nein              |                            //
//                 | Read(0) != 0xFFFF |------------------>|                            //
//                  \_________________/                    |                            //
//                           |  ja                         |                            //
//                           |                             |                            //
//                   ________|________            _________|__________                  //
//                   Enter Application            Enter Program Loader                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked)) __attribute__((section(".init3"))) void
PreStart(void)
{
    UInt08 mcusr = MCUSR;
    MCUSR = 0;
    UInt16 mode = prgLdrMode;
    wdt_enable(WDTO_2S);
    if (mcusr & BV(PORF))
    {   mode = !CFG_PGMLDR_MARKER;
    }
    else
    {   if (mcusr & BV(EXTRF))
        {   mode = CFG_PGMLDR_MARKER;
        }
        else
        {   if (!(mcusr & BV(WDRF)))
            {   mode = !CFG_PGMLDR_MARKER;
            }
        }
    }
    if (mode != CFG_PGMLDR_MARKER && pgm_read_word_near(0) != 0xFFFF) GOTO(0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART0_RX_vect)
{
    UInt08 byte = UDR0;
    UInt08 tail = busRxQueueTail+1;
    if (tail == CARDINALITY(busRxQueue)) tail = 0;
    if (tail == busRxQueueHead) return;
    busRxQueue[busRxQueueTail] = byte;
    busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART1_RX_vect)
{
    UInt08 status = UCSR1A;
    UInt08 byte   = UDR1;
    if (status & (BV(FE1)|BV(DOR1))) return;
    UInt08 tail = serRxQueueTail+1;
    if (tail == CARDINALITY(serRxQueue)) tail = 0;
    if (tail == serRxQueueHead) return;
    serRxQueue[serRxQueueTail] = byte;
    serRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(__vector_default) __attribute__((naked));
ISR(__vector_default)
{
    // Programmlader Einstiegspunkt
    Restart();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Restart(void)
{
    doNext = DO_RESTART;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterAppl(void)
{
    doNext = DO_ENTER_APPL;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmitByte(UInt08 byte)
{
    loop_until_bit_is_set(UCSR0A, UDRE0);
    UDR0 = byte;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceiveByte(UInt08* byte)
{
    if (busRxQueueHead == busRxQueueTail)
    {   *byte = '\0';
        return false;
    }
    Atomic
    {   *byte = busRxQueue[busRxQueueHead];
        busRxQueueHead++;
        if (busRxQueueHead == CARDINALITY(busRxQueue)) busRxQueueHead = 0;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SerTransmitByte(UInt08 byte)
{
    loop_until_bit_is_set(UCSR1A, UDRE1);
    UDR1 = byte;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_SerReceiveByte(UInt08* byte)
{
    if (serRxQueueHead == serRxQueueTail)
    {   *byte = '\0';
        return false;
    }
    Atomic
    {   *byte = serRxQueue[serRxQueueHead];
        serRxQueueHead++;
        if (serRxQueueHead == CARDINALITY(serRxQueue)) serRxQueueHead = 0;
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(UInt16 adr, void* mem, UInt16 size)
{
    eeprom_read_block(mem, (void*)adr, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(UInt16 adr, const void* mem, UInt16 size)
{
    eeprom_update_block(mem, (void*)adr, size);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_DisplayPrint(const char* text)
{
    UInt08 i;

    // Zeile 1
    for (i=0; i<16; i++)
    {   DisplayWrite(false, 0x80 + 0x00 + i);
        DisplayWrite(true, *text);
        text++;
    }
    // Zeile 2
    for (i=0; i<16; i++)
    {   DisplayWrite(false, 0x80 + 0x40 + i);
        DisplayWrite(true, *text);
        text++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    // Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(A, _T_,    // PA.0:            : ADC Kanal 0 (L2IMon)
                _T_,    // PA.1:            : ADC Kanal 1 (L1IMon)
                _T_,    // PA.2:            : ADC Kanal 2 (L2UMon)
                _T_,    // PA.3:            : ADC Kanal 3 (L1UMon)
                _H_,    // PA.4: XCP0       : XPort ???, H = ???
                _L_,    // PA.5: XPRST      : XPort Reset, L = Reset
                _P_,    // PA.6: PIR_SENSOR : PIR-Bewegungssensor, H = Bewegung
                _L_);   // PA.7: BUS_ACT_LED: Cy4NET-Aktivitts-LED, L = Aus
    ConfPort(B, _L_,    // PB.0:            : -+
                _L_,    // PB.1:            :  |
                _L_,    // PB.2:            :  | Datenleitungen
                _L_,    // PB.3:            :  | D0 - D7 der
                _L_,    // PB.4:            :  | Anzeige
                _L_,    // PB.5: [MOSI]     :  |
                _L_,    // PB.6: [MISO]     :  |
                _L_);   // PB.7: [SCK]      : -+
    ConfPort(C, _P_,    // PC.0: [SCL]
                _P_,    // PC.1: [SDA]
                _T_,    // PC.2: [TCK]      : PUSHBUTTON2: Drucktaster 2, L = Gedrckt
                _T_,    // PC.3: [TMS]      : PUSHBUTTON1: Drucktaster 1, L = Gedrckt
                _H_,    // PC.4: [TDO]      : DISPLAY_BL : Anzeige - Beleuchtung, L = Aus
                _L_,    // PC.5: [TDI]      : DISPLAY_EN : Anzeige - "Enable", L = Aus
                _L_,    // PC.6: DISPLAY_RW : Anzeige - "Read/Write", L = Schreiben
                _L_);   // PC.7: DISPLAY_RS : Anzeige - "Register Select", L = Auswahl 0
    ConfPort(D, _P_,    // PD.0: [RXD0]
                _P_,    // PD.1: [TXD0]
                _P_,    // PD.2: [RXD1]
                _P_,    // PD.3: [TXD1]
                _H_,    // PD.4: RELAY_SET  : Relaisstellungswahl, L = Stell. 1
                _L_,    // PD.5: RELAY2_EN  : Relais 2 schalten, L = Aus
                _L_,    // PD.6: RELAY1_EN  : Relais 1 schalten, L = Aus
                _L_);   // PD.7:            : Piezo Signalgeber (Frequenz ber Timer 2)
    // Interrupt-Vektortabelle in den Bootloader-Bereich verlegen
    asm volatile(    "push  R22      \n\t"
                     "IN    R22, %0  \n\t"
                     "SBR   R22, %1  \n\t"
                     "OUT   %0, R22; \n\t"
                     "CBR   R22, %1  \n\t"
                     "SBR   R22, %2  \n\t"
                     "OUT   %0, r22; \n\t"
                     "POP   R22      \n\t"
                    :
                    :"M" (_SFR_IO_ADDR(MCUCR)), "i" (BV(IVCE)), "i" (BV(IVSEL))
                    :"r22");
    Freeze100us();
    // 16-Bit Timer 1 ohne IRQ zhlen lassen (F_CPU1024) --------------------------------
    TCCR1B = BV(CS12) | BV(CS10);
    // USART 0 19200 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC ------------------
    UCSR0A = 0;
    UCSR0B = BV(RXCIE0) | BV(RXEN0) | BV(TXEN0);
    UBRR0  = F_CPU/16L/19200-1;
    // USART 1 57600 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC ------------------
    UCSR1A = 0;
    UCSR1B = BV(RXCIE1) | BV(RXEN1) | BV(TXEN1);
    UBRR1  = F_CPU/16L/57600-1;
	// -----------------------------------------------------------------------------------
    EnableInterrupts();
    //DisplayEnable();
    Freeze100us();
    DisplayEnable();
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_KeepRunning(void)
{
    return doNext == DO_CONTINUE;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
    wdt_reset();
    UInt16 ticks = TIMERTICKS/12;
    ChgPin(SYSTEM_LED, !(ticks & 0x0320));
    ProceedRelaySwitchOff();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
    switch (doNext)
    {   case DO_RESTART:    Restart();
        case DO_ENTER_APPL: EnterAppl();
        default:            doNext = DO_RESTART;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
